#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <stdbool.h>

#define NUM_ADMIN 4

#define BRISANJE 1
#define ZAPOSLJAVANJE 2
#define DOD_PROJ 3
#define OTPUSTANJE 4

#define BENCH_LIMIT 5

typedef struct programer{
    int id;
    int brProj;
    struct programer *next;
}Programer;

Programer *newProgramer(int id, int brProj){
    Programer *new = malloc(sizeof(Programer));
    new->id = id;
    new->brProj = brProj;
    new->next = NULL;
    return new;
}

Programer* dodajNovog(Programer *head, int id){

    Programer *novi = newProgramer(id, 0);

    novi->next = head;

    return novi;
}

void obrisiProgramera(Programer **head, int id){
    if(*head == NULL){
        return;
    }
    else{
        if((*head)->id == id){
            //TODO fix this
            Programer *temp = (*head)->next;
            *head = temp;
            free(*head);
        }
        else{
            obrisiProgramera(&((*head)->next), id);
        }
    }
}

void prebrojBench(Programer *head){
    int benchCount = 0;
    while(head != NULL){
        if(head->brProj == 0)
            benchCount++;
        head = head->next;
    }

    if(benchCount > BENCH_LIMIT)
        printf("U firmi ima vise od %d programera bez projekta\n", BENCH_LIMIT);
    else
        printf("U firmi je stanje kako treba, nema vise od %d programera na bench-u\n", BENCH_LIMIT);
}

void promeniProjekat(Programer **head, int id, int brProj){

    Programer *zaMenjanje = *head;

    while(zaMenjanje->id != id && zaMenjanje->next != NULL){
        zaMenjanje = zaMenjanje->next;
    }

    if(zaMenjanje->id != id && zaMenjanje->next == NULL){
        return;
    }

    zaMenjanje->brProj = brProj;
}

bool pretraga = false;
bool brisanje = false;
bool zaposljavanje = false;
bool postavljanjeProjekta = false;

pthread_mutex_t brojanjeLock;
pthread_mutex_t brisanjeLock;
pthread_mutex_t dodavanjeLock;

pthread_cond_t brisanjeUToku;
pthread_cond_t dodavanjeUToku;
pthread_cond_t brojanjeUToku;

Programer *head = NULL;
int lastID = 0;


void* pretrazi(void *tid){

    long id = (long) tid;

    sleep(rand()%5);
    printf("(%ld) Ja pretrazujem listu\n", id);

    pthread_mutex_lock(&brojanjeLock);

    pthread_mutex_lock(&brisanjeLock);
    while(brisanje == true){
        pthread_cond_wait(&brisanjeUToku, &brisanjeLock);
    }

    pretraga = true;

    sleep(rand()%2+1);
    prebrojBench(head);

    pretraga = false;

    pthread_mutex_unlock(&brisanjeLock);
    pthread_cond_broadcast(&brojanjeUToku);
    pthread_mutex_unlock(&brojanjeLock);

}


void* zaposljavaj(void *tid){

    long id = (long) tid;

    sleep(rand()%5);
    printf("(%ld) Ja zaposljavam novog\n", id);

    pthread_mutex_lock(&dodavanjeLock);
    while(zaposljavanje == true){
        pthread_cond_wait(&dodavanjeUToku, &dodavanjeLock);
    }

    pthread_mutex_lock(&brisanjeLock);
    while(brisanje == true){
        pthread_cond_wait(&brisanjeUToku, &brisanjeLock);
    }

    sleep(rand()%3+1);
    zaposljavanje = true;

    head = dodajNovog(head, lastID);
    lastID++;

    printf("Zaposlio sam programera sa IDem: %d\n", lastID);

    pthread_mutex_unlock(&brisanjeLock);
    zaposljavanje = false;
    pthread_cond_broadcast(&dodavanjeUToku);
    pthread_mutex_unlock(&dodavanjeLock);

}


void* postaviProjekat(void *tid){

    long id = (long) tid;

    sleep(rand()%5);
    printf("(%ld) Ja dodeljujem projekte\n", id);

    pthread_mutex_lock(&dodavanjeLock);
    while(zaposljavanje == true){
        pthread_cond_wait(&dodavanjeUToku, &dodavanjeLock);
    }

    pthread_mutex_lock(&brisanjeLock);
    while(brisanje == true){
        pthread_cond_wait(&brisanjeUToku, &brisanjeLock);
    }

    pthread_mutex_lock(&brojanjeLock);
    while(pretraga == true){
        pthread_cond_wait(&brojanjeUToku, &brojanjeLock);
    }

    sleep(rand()%1+1);

    int idZaIzmenu = rand() % lastID;
    promeniProjekat(&head, idZaIzmenu, rand()%20);

    printf("Promenio sam projekat programeru %d\n", idZaIzmenu);

    pthread_mutex_unlock(&brisanjeLock);
    pthread_mutex_unlock(&dodavanjeLock);
    pthread_mutex_unlock(&brojanjeLock);

}


void* otpustiProgramera(void *tid){

    long id = (long) tid;

    sleep(rand()%5);
    printf("(%ld) Ja otpustam programere\n", id);

    pthread_mutex_lock(&brisanjeLock);
    while(brisanje == true){
        pthread_cond_wait(&brisanjeUToku, &brisanjeLock);
    }

    pthread_mutex_lock(&dodavanjeLock);
    while(zaposljavanje == true){
        pthread_cond_wait(&dodavanjeUToku, &dodavanjeLock);
    }

    pthread_mutex_lock(&brojanjeLock);
    while(pretraga == true){
        pthread_cond_wait(&brojanjeUToku, &brojanjeLock);
    }
    
    brisanje = true;

    sleep(rand()%6+1);

    int idZaBrisanje = rand() % lastID;
    obrisiProgramera(&head, idZaBrisanje);

    printf("Otpustio sam programera %d\n", idZaBrisanje);

    brisanje = false;

    pthread_mutex_unlock(&brisanjeLock);
    pthread_cond_broadcast(&brisanjeUToku);
    pthread_mutex_unlock(&dodavanjeLock);
    pthread_mutex_unlock(&brojanjeLock);

}


int main(){

    pthread_mutex_init(&brojanjeLock, NULL);
    pthread_mutex_init(&brisanjeLock, NULL);
    pthread_mutex_init(&dodavanjeLock, NULL);

    pthread_cond_init(&brojanjeUToku, NULL);
    pthread_cond_init(&brisanjeUToku, NULL);
    pthread_cond_init(&dodavanjeUToku, NULL);

    pthread_t administratori[30];

    long i = 0;

    //zaposljavanje
    for(i = 0 ; i < 15; i++){
        pthread_create(&administratori[i], NULL, zaposljavaj, (void*) i);
    }

    //postavljanje projekata
    for(i ; i < 23 ; i++){
        pthread_create(&administratori[i], NULL, postaviProjekat, (void*) i);
    }

    //prebrojavanje bencha
    for(i; i < 25 ; i++){
        pthread_create(&administratori[i], NULL, pretrazi, (void*) i);
    }
    
    // //otpustanje
    // for(i ; i < 30 ; i++){
    //     pthread_create(&administratori[i], NULL, otpustiProgramera, (void*) i);
    // }

    for(long i = 0 ; i < 30 ; i++){
        pthread_join(administratori[i], NULL);
    }

}