

#include<stdio.h>
#include<stdlib.h>
#include<pthread.h>

#define BROJ_ZAPOSLAVANJE 10
#define BROJ_OTPUSTANJE 2
#define BROJ_NEDODELJENIH_PROJ 3
#define BROJ_POSTAVLJANJA_PROJ 2

#define MAKS_PROGRAMERA 100
#define MAKS_PROJEKATA 60

typedef struct 
{
	int id;
	int brojProjekta;
} programer_t;

typedef struct lista_t
{
	programer_t programer;
	struct lista_t *sledeci;
} lista_t;

int brisanje = 0;
int dodavanje = 0;
int pretraga = 0;

int brojElemenata = 0;
lista_t *p = NULL;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

lista_t *alociraj(int id)
{
	lista_t *el = (lista_t *)malloc(sizeof(lista_t));
	el->programer.id = id;
	el->programer.brojProjekta = 0;
	return el;
}

void dodajProgramera(int id)
{
	lista_t *novi = alociraj(id);
	if (p == NULL)
	{
		p = novi;
		return;
	}
	
	novi->sledeci = p;
	p = novi;
	brojElemenata++;
}

void obrisiProgramera(int id)
{
	if (p == NULL)
		return;
	
	lista_t *prethodni = NULL;
	lista_t *tren = p;
	
	while (tren != NULL && tren->programer.id != id)
		tren = tren->sledeci;
	
	if (prethodni == NULL)
	{
		lista_t *pom = p;
		p = p->sledeci;
		free(pom);
		return;
	}
	
	lista_t *pom = tren;
	prethodni->sledeci = tren->sledeci;
	free(pom);
}

programer_t *pronadjiProgramera(int id)
{
	lista_t *pom = p;
	while (pom != NULL)
	{
		if (pom->programer.id == id)
			return &pom->programer;
		pom = pom->sledeci;
	}
	
	return NULL;
}

int pretraziBrojNedodeljenihProjekata()
{
	lista_t *l = p;
	int br = 0;
	while (l != NULL)
	{
		if (l->programer.brojProjekta == 0)
			br++;
		l = l->sledeci;
	}
	
	return br;
}

void *daLiImaKojiNeRadeProjekat(void *arg)
{
	long id = (long)arg;
	
	pthread_mutex_lock(&mutex);
	while (brisanje > 0)
		pthread_cond_wait(&cond, &mutex);
		
	pretraga++;
	pthread_mutex_unlock(&mutex);
	
	int brojProgramera = rand() % (MAKS_PROGRAMERA / 10);
	printf("Administrator %ld proverava da li radi %d ili vise programera koji nemaju dodeljen projekat\n", id, brojProgramera);
	int brojNedodeljenihProj = pretraziBrojNedodeljenihProjekata();
	if (brojNedodeljenihProj >= brojProgramera)
		printf("U kompaniji radi %d ili vise programera koji nemaju dodeljen projekat\n", brojProgramera);
	else
		printf("U kompaniji nema %d ili vise programera koji nemaju dodeljen projekat\n", brojProgramera);
	
	pthread_mutex_lock(&mutex);
	pretraga--;
	pthread_mutex_unlock(&mutex);
	pthread_cond_broadcast(&cond);
	
	pthread_exit(NULL);
}

void *zaposliNovog(void *arg)
{
	long id = (long)arg;
	pthread_mutex_lock(&mutex);
	while (brisanje > 0 || dodavanje > 0)
		pthread_cond_wait(&cond, &mutex);
		
	dodavanje++;
	pthread_mutex_unlock(&mutex);
	
	// zaposlavanje novog programera
	programer_t *prog = NULL;
	int progId;
	do
	{
		progId = rand() % MAKS_PROGRAMERA;
		prog = pronadjiProgramera(progId);
	} while (prog != NULL);
	
	printf("Administrator %ld dodaje programera sa id-em %d\n", id, progId);
	dodajProgramera(progId);
	printf("Administrator %ld je dodao novog programera sa id-em %d\n", id, progId);
	
	pthread_mutex_lock(&mutex);
	dodavanje--;
	pthread_mutex_unlock(&mutex);
	pthread_cond_broadcast(&cond);
	
	pthread_exit(NULL);
}

void *postaviNaProjekat(void *arg)
{
	long id = (long)arg;
	pthread_mutex_lock(&mutex);
	while (brisanje > 0)
		pthread_cond_wait(&cond, &mutex);
		
	pretraga++;
	pthread_mutex_unlock(&mutex);
		
	// postavljanje programera na projekta
	programer_t *programer = NULL;
	
	do
	{
		int progId = rand() % MAKS_PROGRAMERA;
		programer = pronadjiProgramera(progId);
	} while (programer == NULL);
	int projekat = rand() % MAKS_PROJEKATA + 1;
	printf("Administrator %ld postavlja programera %d na projekat %d\n", id, programer->id, projekat);
	programer->brojProjekta = projekat;
	printf("Administrator %ld je postavio programera %d na projekat %d\n", id, programer->id, projekat);
	
	pthread_mutex_lock(&mutex);
	pretraga--;
	pthread_mutex_unlock(&mutex);
	pthread_cond_broadcast(&cond);
	
	pthread_exit(NULL);
}

void *otpustiProgramera(void *arg)
{
	long id = (long)arg;
	
	pthread_mutex_lock(&mutex);
	while (brisanje > 0 || dodavanje > 0 || pretraga > 0)
		pthread_cond_wait(&cond, &mutex);
		
	brisanje++;
	pthread_mutex_unlock(&mutex);
	
	// otpustanje programera
	programer_t *programer = NULL;
	do
	{
		int progId = rand() % MAKS_PROGRAMERA;
		programer = pronadjiProgramera(progId);
	} while (programer == NULL);
	
	printf("Administrator %ld otpusta programera %d\n", id, programer->id);
	obrisiProgramera(programer->id);
	printf("Administrator %ld je otpustio programera %d\n", id, programer->id);
	
	pthread_mutex_lock(&mutex);
	brisanje--;
	pthread_mutex_unlock(&mutex);
	pthread_cond_broadcast(&cond);
	
	pthread_exit(NULL);
}

int main()
{
	srand(time(NULL));
	
	pthread_t B[BROJ_NEDODELJENIH_PROJ];
	pthread_t Z[BROJ_ZAPOSLAVANJE];
	pthread_t D[BROJ_POSTAVLJANJA_PROJ];
	pthread_t O[BROJ_OTPUSTANJE];
	
	long ukId = 0l;
	for (int i = 0; i < BROJ_ZAPOSLAVANJE; i++)
		pthread_create(Z + i, NULL, zaposliNovog, (void *)ukId++);
		
	for (int i = 0; i < BROJ_POSTAVLJANJA_PROJ; i++)
		pthread_create(D + i, NULL, postaviNaProjekat, (void *)ukId++);
		
	for (int i = 0; i < BROJ_NEDODELJENIH_PROJ; i++)
		pthread_create(B + i, NULL, daLiImaKojiNeRadeProjekat, (void *)ukId++);
		
	for (int i = 0; i < BROJ_OTPUSTANJE; i++)
		pthread_create(O + i, NULL, otpustiProgramera, (void *)ukId++);
	
	pthread_exit(NULL);
	
	return 0;
}
