#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<pthread.h>

#define BROJ_ADMINISTRATORA 5

typedef struct programer{
    int id_programera, broj_projekata_programera;
    struct programer *sledeci;   
}Programer;

Programer *lista = NULL;

int broj_programera = 20;

int pretraga = 0;
int dodavanje = 0;
int izmena = 0;
int brisanje = 0;

pthread_mutex_t mutex;
pthread_cond_t cond;

void daLiImaViseOnihSa0Projekata(int x){
    Programer *p = lista;
    int counter = 0;
    if(p != NULL && lista->broj_projekata_programera == 0){
        counter ++;
    }
    while(p->sledeci != NULL){
        p = p->sledeci;
        if(p->broj_projekata_programera == 0)
            counter ++;
    }
    
    if(counter > x){
        printf("U firmi je vise programera od %d sa 0 projekata - tacnije: %d \n ",x,counter);
    }
    else{
        printf("U firmi je manje ili jednako programera od %d sa 0 projekata - tacnije: %d \n ",x,counter);
    }
}

void dodavanjeNaPocetakListe(){
    Programer *novi = (Programer *) malloc(sizeof(Programer));
    novi->id_programera = broj_programera;
    broj_programera ++;
    novi->broj_projekata_programera = 0;
    novi->sledeci = lista;
    lista = novi;
}

Programer * pronadjiProgramera(int id){
    Programer *p = lista;
    int counter = 0;
    if(p != NULL && lista->id_programera == id){
        return p;
    }
    while(p->sledeci != NULL){
        p = p->sledeci;
        if(p->id_programera == id)
            return p;
    }
    return NULL;
}

Programer * pronadjiProgrameraIVratiPrethodnog(int id){
    Programer *prethodni = NULL;
    Programer *p = lista;
    int counter = 0;
    if(p != NULL && lista->id_programera == id){
        return NULL;
    }
    while(p->sledeci != NULL){
        prethodni = p;
        p = p->sledeci;

        if(p->id_programera == id)
            return prethodni;
    }
    return NULL;
}

void* administracija(void *arg){
    long id = (long) arg;
    int ja_cu_da_radim = ((rand()%4) + 1);
    if(ja_cu_da_radim == 1){
        // B X odgovaraju da li u kompaniji trenutno radi X ili vise programera koji nemaju dodeljeni projekat
        printf("Admin %ld ce vrsiti operaciju B\n", id);
        pthread_mutex_lock(&mutex);
        int X = rand() % broj_programera;
        while(brisanje > 0){
            printf("\t Admin %ld cekam pristup\n",id);
            pthread_cond_wait(&cond, &mutex);
        }
        pretraga++;
        pthread_mutex_unlock(&mutex);
        daLiImaViseOnihSa0Projekata(X);

        pthread_mutex_lock(&mutex);
        pretraga --;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);

        printf("Admin %ld je zavrsio operaciju B\n", id);
    }
    else if(ja_cu_da_radim == 2){
        // Z - запошљавају новог програмера;
        printf("Admin %ld ce vrsiti operaciju Z\n",id);
        pthread_mutex_lock(&mutex);
        while(brisanje > 0 || dodavanje >0){
            printf("\t Admin %ld cekam pristup\n",id);
            pthread_cond_wait(&cond, &mutex);
        }
        dodavanje ++;
        pthread_mutex_unlock(&mutex);

        dodavanjeNaPocetakListe();

        pthread_mutex_lock(&mutex);
        dodavanje --;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);
        
        printf("Admin %ld je zavrsio operaciju Z - dodao je novog programera\n",id);
    }
    else if(ja_cu_da_radim == 3){
        // D ID P - постављају програмера ID на пројекат P;
        printf("Admin %ld ce vrsiti operaciju D\n",id);
        pthread_mutex_lock(&mutex);
        while(brisanje > 0 || izmena > 0 || pretraga > 0){
            printf("\t Admin %ld cekam pristup\n",id);
            pthread_cond_wait(&cond, &mutex);
        }
        izmena ++;
        int dodacu_programeru = rand() % broj_programera;
        pthread_mutex_unlock(&mutex);

        Programer *p = pronadjiProgramera(dodacu_programeru);
        if(p == NULL){
            printf("Nepostojeci programer\n");
            return 0;
        }
        p->broj_projekata_programera += 1;

        pthread_mutex_lock(&mutex);
        izmena --;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);

        printf("Admin %ld je zavrsio operaciju D - programeru %d je dodao nov projekat \n",id, dodacu_programeru);

    }
    else{
        // O Id otpustaju programera ID
        printf("Admin %ld ce vrsiti operaciju O\n",id);
        pthread_mutex_lock(&mutex);
        while(brisanje > 0 || pretraga >0 || izmena > 0 || dodavanje > 0){
            printf("\t Admin %ld cekam pristup\n",id);
            pthread_cond_wait(&cond, &mutex);
        }
        brisanje ++;
        int obrisacu = rand() % broj_programera;
        pthread_mutex_unlock(&mutex);

        Programer *p = pronadjiProgrameraIVratiPrethodnog(obrisacu);
        if(p == NULL){
            printf("Nepostojeci programer\n");
            return 0;
        }
        Programer *sledeci = p->sledeci;
        p->sledeci = sledeci->sledeci;
        free(sledeci);

        pthread_mutex_lock(&mutex);
        brisanje --;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);

        printf("Admin %ld ce vrsiti operaciju O nad programerom %d\n",id, obrisacu);

    }
}

int main(int argc, char** argv){
    for(int i=0;i<=broj_programera;i++){
        Programer *novi = (Programer *) malloc(sizeof(Programer));
        novi->id_programera = i;
        if(rand() % 2 == 0)
            novi->broj_projekata_programera = (rand() % 10 + 1);
        else
            novi->broj_projekata_programera = 0;
        novi->sledeci = NULL;
        
        if(lista == NULL)
            lista = novi;
        else{
            Programer *p = lista;
            while(p->sledeci != NULL){
                p = p->sledeci;
            }
            p->sledeci = novi;
        }
    }

    printf("Stanje pre operacija\n");
    Programer *p = lista;
    while(p->sledeci != NULL){
        printf("%d - %d \n",p->id_programera, p->broj_projekata_programera);
        p = p->sledeci;
    }
    printf("===============================\n");
    

    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);

    pthread_t niti[BROJ_ADMINISTRATORA];
    srand(time(NULL));
    for(long i=0;i<BROJ_ADMINISTRATORA;i++){
        pthread_create(niti+i, NULL, administracija, (void *)i);
    }

    for(int i=0;i<BROJ_ADMINISTRATORA;i++){
        pthread_join(niti[i], NULL);
    }

    // dodavanjeNaPocetakListe();
    // dodavanjeNaPocetakListe();
    printf("===============================\n");
    printf("Stanje posle operacija\n");
    p = lista;
    while(p->sledeci != NULL){
        printf("%d - %d \n",p->id_programera, p->broj_projekata_programera);
        p = p->sledeci;
    }


    return 0;
}