#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<pthread.h>

#define BROJ_ZAPOSLENIH 10
#define BROJ_PROGRAMERA 50

typedef struct programer{
    int identifikacioniBroj;
    int brProjekata;
    struct programer *sledeci;
}Programer;

Programer *lista;

pthread_mutex_t mutex;
pthread_cond_t cond;

int brojPretrage = 0;
int brojUnosa = 0;
int brojBrisanja = 0;

int id = 0;
int duzinaListe = 0;

//pretrazivanje = 0
//unos = 1
//brisanje = 2

void proveri(Programer *lista, int broj)
{
    Programer *glava = lista;
    int br = 0;
    
    while(glava)
    {
        if(glava->brProjekata == 0)
        {
            br++;
        }

        glava = glava->sledeci;
    }
    
    if(br >= broj)
        printf("Pronasao sam %d programera sa 0 projekta\n", br);
    
    else
        printf("Postoji manje programera sa 0 projekta nego sto je trazeno. Broj nadjenih je %d\n", br);

}

void pretrazi(Programer *lista, int broj)
{
    Programer *glava = lista;
    
    while(glava)
    {
        if(glava->identifikacioniBroj == broj)
        {
            printf("Pronasao sam programera sa id-jem %d\n", broj);
            return;
        }

        glava = glava->sledeci;
    }

    printf("ne postoji programer sa id-jem %d\n", broj);
}

Programer *dodaj(Programer *glava, int broj)
{
    Programer *novi = (Programer *)malloc(sizeof(Programer));

    novi->identifikacioniBroj = broj;
    novi->brProjekata = 0;
    
    if(glava == NULL) //lista programera je prazna
        novi->sledeci = NULL;
    
    else //lista programera nije prazna
    {
        novi->sledeci = glava;
        glava = novi;
    }

    printf("Dodao sam programera\n");
    return novi;

}

void brisi(Programer *glava, int id)
{
    //potrebni su nam trenutni i prethodni
    Programer *gl = glava;
    Programer *trenutni = glava;
    Programer *prethodni = NULL;

    if(trenutni == NULL)
    {
        printf("Lista je prazna, pa je nemoguce brisanje\n");
        return;
    }

    if(trenutni && trenutni->identifikacioniBroj == id) //brisem prvi element liste
    {
        if(trenutni->sledeci)
        {
            gl = trenutni->sledeci;
            return;
        }
        else
            return;
    }

    trenutni = gl;

    while(trenutni)
    {
        if(trenutni->identifikacioniBroj == id)
        {
            if(trenutni->sledeci)
                prethodni->sledeci = trenutni->sledeci;
            else
                prethodni->sledeci = NULL;
        }
        prethodni = trenutni;
        trenutni = trenutni->sledeci;
    }
}

void *radi(void *arg)
{
    long id = (long)arg;

    int posao = rand() % 5;

    if(posao == 0) //pretrazivanje, uvek osim ako se trenutno radi brisanje
    {

        //int p = rand() % duzinaListe;

        pthread_mutex_lock(&mutex);
        while(brojBrisanja != 0)
            pthread_cond_wait(&cond, &mutex);

        brojPretrage++;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld pretrazuje programere\n", id);
        //funkcija za pretragu ovde ide

        pthread_mutex_lock(&mutex);
        brojPretrage--;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld je zavrsio pretragu programera\n", id);

    }
    else if(posao == 1) //unos, uvek osim ako se trenutno radi brisanje ili neko drugi radi unos isto
    {
        pthread_mutex_lock(&mutex);
        while(brojBrisanja != 0 || brojUnosa != 0)
            pthread_cond_wait(&cond, &mutex);
        
        id++;
        brojUnosa++;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld unosi programera\n", id);
        //funkcija za unos ovde ide
        dodaj(lista, id);

        pthread_mutex_lock(&mutex);
        brojUnosa--;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld je zavrsio unos programera\n", id);
    }
    else //brisanje, uvek osim ako se trenutno radi pretrazivanje, unos ili neko drugo brisanje
    {
        pthread_mutex_lock(&mutex);
        while(brojBrisanja != 0 || brojUnosa != 0 || brojPretrage != 0)
            pthread_cond_wait(&cond, &mutex);

        brojBrisanja++;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld brise programera\n", id);
        //funkcija za brisanje ovde ide
        //brisi(lista, rand() % duzinaListe);

        pthread_mutex_lock(&mutex);
        brojBrisanja--;
        pthread_mutex_unlock(&mutex);

        printf("Administrator %ld je zavrsio brisanje programera\n", id);
    }

}

int main()
{

    lista = (Programer *)malloc(BROJ_PROGRAMERA*sizeof(Programer));

    pthread_t zaposleni[BROJ_ZAPOSLENIH];
    
    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);

    for(long i=0;i<BROJ_ZAPOSLENIH;i++)
    {
        pthread_create(zaposleni + i, NULL, radi, (void *)i);
    }

    for(int i=0;i<BROJ_ZAPOSLENIH;i++)
    {
        pthread_join(zaposleni[i], NULL);
    }

    return 0;

}