//Aleksandra Stanic 82/2020
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

#define ADMINISTRACIJA 15

#define B_X     0                       // pretraga
#define Z       1                       // zaposljavanje
#define D_ID_P  2                       // dodavanje projekta
#define O_ID    3                       // otpustanje

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

typedef struct programer
{
    long id;
    int broj_projekata;                 // 0 - ako ne radi na projektu
    struct programer *sledeci;
}Programer;

Programer *p;
long id_programera = 0;

void *zaposljavanje(long id, int broj_projekata)
{
    Programer *temp = (Programer*)malloc(sizeof(Programer));
    temp->id = id;
    temp->broj_projekata = broj_projekata;
    temp->sledeci = p;
    p = temp;
    printf("ZAPOSLJAVANJE - Programer %ld.\n", id);
}

void *otpustanje(long id)
{
    Programer *temp = (Programer*)malloc(sizeof(Programer));
    temp = p;

    if(temp == NULL)
    {
        printf("GRESKA - Jos uvek nema zaposljenih porogramera.\n");
    }
     
    while(temp->sledeci->id != id)
        temp = temp->sledeci;
    
    Programer *pom = temp->sledeci;
    temp->sledeci = pom->sledeci;
    free(pom);
    printf("OTPUSTANJE - Programer %ld.\n", id);
}

void *angazovanje(long id)
{
    Programer *temp = (Programer*)malloc(sizeof(Programer));
    temp = p;

    if(temp == NULL)
    {
        printf("GRESKA - Jos uvek nema zaposljenih porogramera.\n");
    }
     
    while(temp->id != id)
        temp = temp->sledeci;
    temp->id++;

    printf("DODAVANJE NA PROJEKAT - Programer %ld.\n", id);
}

void provera(int x)
{
    int brojac = 0;

    Programer *temp = (Programer*)malloc(sizeof(Programer));
    temp = p;

    if(temp == NULL)
    {
        printf("GRESKA - Jos uvek nema zaposljenih porogramera.\n");
    }
     
    while(temp->broj_projekata == 0)
    {
        brojac++;
        temp = temp->sledeci;
    }
    
    if(brojac > x) printf("PROVERA - Trenutno ima vise od %d programera koji nemaju dodeljeni projekat.\n", x);
    else printf("PROVERA - Trenutno inema vise od %d programera koji nemaju dodeljeni projekat.\n", x);
}

int dodavanje = 0;
int brisanje = 0;
int aktivnost = 0;
int pretraga = 0;


void *work(void *arg)
{
    long id = (long)arg;
    sleep(2);
    printf("ADMINISTRATOR %ld.\n", id);

    //sta ce da radi administrator
    int posao = rand() % 4; 

    if(posao == 0) // B X
    {
        pthread_mutex_lock(&mutex);
        int x = rand() % 10;

        //pretrazuje samo ako nema brisanja
        while(brisanje == 1) 
            pthread_cond_wait(&cond, &mutex);

        pretraga = 1;
        provera(x);
        pthread_mutex_unlock(&mutex);

        sleep(5);

        pthread_mutex_lock(&mutex);
        pretraga = 0;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);
    }

    if(posao == 1) // Z
    {
        pthread_mutex_lock(&mutex);
        id_programera++;

        //dodaje samo ako niko ne brise ili ako niko drugi ne dodaje
        while(brisanje == 1 || dodavanje == 1)
            pthread_cond_wait(&cond, &mutex);

        dodavanje = 1;
        zaposljavanje(id_programera, 0);
        pthread_mutex_unlock(&mutex);

        sleep(5);

        pthread_mutex_lock(&mutex);
        dodavanje = 0;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);
    }

    if(posao == 2) // D ID P
    {
        pthread_mutex_lock(&mutex);

        // dodaje na projekat ako niko ne brise
        // obezbedjuje se mogucnost da samo jedan u isto vreme moze da angazuje programera
        while(brisanje == 1 || aktivnost == 1)
            pthread_cond_wait(&cond, &mutex);

        aktivnost = 1;
        angazovanje(rand() % id_programera);
        pthread_mutex_unlock(&mutex);

        sleep(5);

        pthread_mutex_lock(&mutex);
        aktivnost = 0;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);
    }

    if(posao == 3) // O ID
    {
        pthread_mutex_lock(&mutex);
        int X = rand() % 10;

        // brise samo ako nema nikakve akcije nad listom
        while(aktivnost == 1 || brisanje == 1 || dodavanje == 1 || pretraga == 1)
            pthread_cond_wait(&cond, &mutex);

        brisanje = 1;
        otpustanje(rand() % id_programera);
        pthread_mutex_unlock(&mutex);

        sleep(5);

        pthread_mutex_lock(&mutex);
        brisanje = 0;
        pthread_cond_broadcast(&cond);
        pthread_mutex_unlock(&mutex);
    }
    //pthread_t(exit);
}

int main()
{
    p = (Programer*)malloc(sizeof(Programer));
    p = NULL;
    pthread_t nit[ADMINISTRACIJA];

    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);

    for (long i = 0; i < ADMINISTRACIJA; i++)
    {
        pthread_create(nit + i, NULL, *work, (void *)i);
    }

    for (long i = 0; i < ADMINISTRACIJA; i++)
    {
        pthread_join(nit[i], NULL);
    }

    pthread_mutex_destroy(&mutex);
    pthread_cond_destroy(&cond);

    return 0;
}