// reset && gcc programeri.c -o programeri -lpthread && ./programeri
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdbool.h>

struct node
{
    int id;
    int proj_id;
    struct node *next;
};

struct cmd
{
    char op;
    int arg1, arg2;
};

struct node *programeri = NULL;
int ids = 0;

pthread_mutex_t protect, delete, common;
pthread_cond_t cdelete, ccommon;

bool dprogress = false;
int otherop = 0;

struct node *novi()
{
    struct node *n = malloc(sizeof(struct node));

    n->next = NULL;

    if (n == NULL)
    {
        fprintf(stderr, "FAILED %d\n", __LINE__);
        exit(EXIT_FAILURE);
    }
}

void dodaj(int id, int proj_id, struct node **_l)
{
    pthread_mutex_lock(&delete);
    if (dprogress)
    {
        pthread_cond_wait(&cdelete, &delete);
    }

    pthread_mutex_lock(&common);
    otherop++;
    pthread_mutex_unlock(&common);

    pthread_mutex_lock(&protect);
    struct node *l = *_l;

    struct node *n = novi();

    n->id = id;
    n->proj_id = proj_id;

    if (l == NULL)
    {
        l = n;
    }
    else
    {
        n->next = l->next;
        l = n;
    }

    *_l = l;

    pthread_mutex_unlock(&protect);

    pthread_mutex_lock(&common);
    otherop--;
    if (otherop == 0)
    {
        pthread_cond_broadcast(&ccommon);
    }
    pthread_mutex_unlock(&common);

    pthread_mutex_unlock(&delete);
}

void brisi(int id, struct node **_l)
{
    pthread_mutex_lock(&delete);
    dprogress = true;

    pthread_mutex_lock(&common);
    while (otherop != 0)
    {
        pthread_cond_wait(&ccommon, &common);
    }

    pthread_mutex_lock(&protect);

    struct node *l = *_l, *prev, *t;

    if (l == NULL)
    {
        fprintf(stderr, "ERROR %d\n", __LINE__);
    }
    else
    {
        for (t = l, prev = NULL; t && t->id != id; prev = t, t = t->next)
            ;
        if (prev == NULL)
        {
            t = l;
            l = l->next;
            free(t);
        }
        else
        {
            prev->next = t->next;
            free(t);
        }
    }

    *_l = l;

    pthread_mutex_unlock(&protect);

    dprogress = false;

    pthread_cond_broadcast(&cdelete);

    pthread_mutex_unlock(&common);
    pthread_mutex_unlock(&delete);
}

struct node *pretrazi_id(int id, struct node *list)
{
    pthread_mutex_lock(&delete);
    if (dprogress)
    {
        pthread_cond_wait(&cdelete, &delete);
    }

    pthread_mutex_lock(&common);
    otherop++;
    pthread_mutex_unlock(&common);

    struct node *t = list;

    for (t; t && t->id != id; t = t->next)
        ;

    pthread_mutex_lock(&common);
    otherop--;
    if (otherop == 0)
    {
        pthread_cond_broadcast(&ccommon);
    }
    pthread_mutex_unlock(&common);

    pthread_mutex_unlock(&delete);
    return t;
}

int pretrazi_proj(struct node *list)
{
    pthread_mutex_lock(&delete);
    if (dprogress)
    {
        pthread_cond_wait(&cdelete, &delete);
    }

    pthread_mutex_lock(&common);
    otherop++;
    pthread_mutex_unlock(&common);

    struct node *t = list;
    int count = 0;

    for (t; t; t = t->next)
    {
        if (t->proj_id == 0)
        {
            count++;
        }
    }

    pthread_mutex_lock(&common);
    otherop--;
    if (otherop == 0)
    {
        pthread_cond_broadcast(&ccommon);
    }
    pthread_mutex_unlock(&common);

    pthread_mutex_unlock(&delete);
    return count;
}

struct node *postavi_na_projekat(int id, int proj_id, struct node *list)
{
    pthread_mutex_lock(&delete);
    if (dprogress)
    {
        pthread_cond_wait(&cdelete, &delete);
    }

    pthread_mutex_lock(&common);
    otherop++;
    pthread_mutex_unlock(&common);

    pthread_mutex_lock(&protect);

    struct node *t = list;

    for (t; t && t->id != id; t = t->next)
        ;

    if (t)
    {
        t->proj_id = proj_id;
    }
    else
    {
        fprintf(stderr, "ERROR %d\n", __LINE__);
    }

    pthread_mutex_unlock(&protect);

    pthread_mutex_lock(&common);
    otherop--;
    if (otherop == 0)
    {
        pthread_cond_broadcast(&ccommon);
    }
    pthread_mutex_unlock(&common);

    pthread_mutex_unlock(&delete);
    return t;
}

void *izvrsi_komandu(void *arg)
{
    struct cmd c = *((struct cmd *)arg);

    if (c.op == 'B')
    {
        int x = c.arg1;

        int r = pretrazi_proj(programeri);
        if (x >= r)
        {
            printf("[%c] U kompaniji ima %d slobodnih programera\n", c.op, r);
        }
        else
        {
            printf("[%c] U kompaniji nema dovoljan broj slobodnih programera\n", c.op);
        }
    }
    else if (c.op == 'Z')
    {
        dodaj(ids, 0, &programeri);

        printf("[%c] Dodat novi programer, id = %d\n", c.op, ids);
        ids++;
    }
    else if (c.op == 'D')
    {
        int id = c.arg1;
        int proj_id = c.arg2;

        postavi_na_projekat(id, proj_id, programeri);

        printf("[%c] Programer id = %d postavljen na projekat id = %d\n", c.op, id, proj_id);
    }
    else if (c.op == 'O')
    {
        int id = c.arg1;

        brisi(id, &programeri);
        printf("[%c] Programer id = %d je otpusten\n", c.op, id);
    }
    else
    {
        fprintf(stderr, "Nepoznata radnja\n");
    }
    return arg;
}

int main(int argc, char **argv)
{
    pthread_mutex_init(&protect, NULL);
    pthread_mutex_init(&delete, NULL);
    pthread_mutex_init(&common, NULL);

    pthread_cond_init(&ccommon, NULL);
    pthread_cond_init(&cdelete, NULL);

    /* U zasebnim nitima izvrsi sljedece komande:
    DODAJ
    DODAJ
    POSTAVI NA PROJEKAT 1
    POSTAVI NA PROJEKAT 1
    DA LI RADE 2 SLOBDNA PROGRAMERA
    OTPUSTI PROGRAMERA
    DA LI RADE 2 SLOBDNA PROGRAMERA
    */
    struct cmd cmds[] = {
        {.op = 'Z'},
        {.op = 'Z'},

        {.op = 'D', .arg1 = 0, .arg2 = 1},
        {.op = 'D', .arg1 = 1, .arg2 = 1},

        {.op = 'B', .arg1 = 2},

        {.op = 'O', .arg1 = 0},
        {.op = 'B', .arg1 = 2}};

    pthread_t threads[sizeof(cmds) / sizeof(cmds[0])];

    for (int i = 0; i < sizeof(threads) / sizeof(threads[0]); i++)
    {
        if (pthread_create(&threads[i], NULL, &izvrsi_komandu, (void *)&cmds[i]) != 0)
        {
            fprintf(stderr, "FAILED %d\n", __LINE__);
        }
    }

    for (int i = 0; i < sizeof(threads) / sizeof(threads[0]); i++)
    {
        if (pthread_join(threads[i], NULL) != 0)
        {
            fprintf(stderr, "FAILED %d\n", __LINE__);
        }
    }

    pthread_mutex_destroy(&protect);
    pthread_mutex_destroy(&delete);
    pthread_mutex_destroy(&common);

    pthread_cond_destroy(&ccommon);
    pthread_cond_destroy(&cdelete);

    for (struct node *prev = NULL; programeri; prev = programeri, programeri = programeri->next)
    {
        if (prev)
        {
            free(prev);
        }
    }
    free(programeri);

    return EXIT_SUCCESS;
}