#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>

typedef struct node{
    int id;
    int projectNum;
    struct node* next;
}Node;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t condE = PTHREAD_COND_INITIALIZER;
pthread_cond_t condS = PTHREAD_COND_INITIALIZER;
pthread_cond_t condA = PTHREAD_COND_INITIALIZER;

int exclusive = 0;   //ekskluzivno pravo, brisanje i postavljanje na projekat
int searching = 0;  //bilo kad sem kad se brise
int adding = 0;     //kad se ne brise ili dodaje neko drugi

Node *p = NULL;

void addNode(Node **p, int id){
    Node *newNode = malloc(sizeof(Node));
    newNode->id = id;
    newNode->projectNum = 0;
    newNode->next = NULL;

    if (*p == NULL) *p = newNode;
    else{
        newNode->next = *p;
        *p = newNode;
    }
}

void createList(Node **p){
    int id;
    scanf("%d", &id);

    while (id != -1){
        addNode(p, id);
        scanf("%d", &id);
    }
}

void printList(Node *p){
    while (p){
        printf("%d %d\n", p->id, p->projectNum);
        p = p->next;
    }
}

void deleteNode(Node **p, int id){
    if ((*p)->id == id){
        Node *temp = *p;
        *p = temp->next;
        free(temp);
    }
    else{
        Node *temp = *p;
        Node *temp_bef = NULL;
        while (temp->id != id){
            temp_bef = temp;
            temp = temp->next;
        }
        temp_bef->next = temp->next;
        free(temp);
    }
}

int withoutProject(Node *p, int x){
    while(p){
        if (p->projectNum == 0) x--;
        p = p->next;
    }

    return x <= 0;
}

void assign(Node *p, int id, int projectNum){
    while (p->id != id) p = p->next;
    p->projectNum = projectNum;
}

#define X 2

void *B(void *i){
    //pretraga, bilo kad osim ako se brise
    long id = (long)i;
    sleep(rand() % 5 + 1);

    printf("%ld: Pokusavam da izvrsim operaciju B\n", id);

    pthread_mutex_lock(&mutex);
    while (exclusive > 0) pthread_cond_wait(&condS, &mutex);

    searching++;
    pthread_mutex_unlock(&mutex);
    
    printf("%ld: Izvrsavam operaciju B...\n", id);
    sleep(rand() % 5 + 1);
    withoutProject(p, X);
    printf("%ld: Izvrsio operaciju B!\n", id);

    pthread_mutex_lock(&mutex);
    searching--;
    pthread_cond_broadcast(&condS);
    if (searching == 0) pthread_cond_broadcast(&condE);
    pthread_mutex_unlock(&mutex);

    pthread_exit(NULL);
}

#define ID 7

void *Z(void *i){
    //dodavanje, kad se ne brise ili neko drugi ne dodaje
    long id = (long)i;
    sleep(rand() % 5 + 1);

    printf("%ld: Pokusavam da izvrsim operaciju Z\n", id);

    pthread_mutex_lock(&mutex);
    while (exclusive > 0 || adding > 0) pthread_cond_wait(&condA, &mutex);

    adding++;
    pthread_mutex_unlock(&mutex);
    
    printf("%ld: Izvrsavam operaciju Z...\n", id);
    sleep(rand() % 5 + 1);
    addNode(&p, ID);
    printf("%ld: Izvrsio operaciju Z!\n", id);

    pthread_mutex_lock(&mutex);
    adding--;
    pthread_cond_broadcast(&condA);
    pthread_cond_broadcast(&condE);
    pthread_mutex_unlock(&mutex);

    pthread_exit(NULL);
}

#define P 2

void *D(void *i){
    //ekskluzivno pravo
    long id = (long)i;
    sleep(rand() % 5 + 1);

    printf("%ld: Pokusavam da izvrsim operaciju D\n", id);

    pthread_mutex_lock(&mutex);
    while (exclusive > 0 || adding > 0 || searching > 0) pthread_cond_wait(&condE, &mutex);

    exclusive++;
    pthread_mutex_unlock(&mutex);
    
    printf("%ld: Izvrsavam operaciju D...\n", id);
    sleep(rand() % 5 + 1);
    assign(p, ID, P);
    printf("%ld: Izvrsio operaciju D!\n", id);

    pthread_mutex_lock(&mutex);
    exclusive--;
    pthread_cond_broadcast(&condE);
    pthread_cond_broadcast(&condA);
    pthread_cond_broadcast(&condS);
    pthread_mutex_unlock(&mutex);

    pthread_exit(NULL);
}


void *O(void *i){
    //brisanje
    long id = (long)i;
    sleep(rand() % 5 + 1);

    printf("%ld: Pokusavam da izvrsim operaciju O\n", id);

    pthread_mutex_lock(&mutex);
    while (exclusive > 0 || searching > 0 || adding > 0) pthread_cond_wait(&condE, &mutex);

    exclusive++;
    pthread_mutex_unlock(&mutex);
    
    printf("%ld: Izvrsavam operaciju O...\n", id);
    sleep(rand() % 5 + 1);
    deleteNode(&p, ID);
    printf("%ld: Izvrsio operaciju O!\n", id);

    pthread_mutex_lock(&mutex);
    exclusive--;
    pthread_cond_broadcast(&condE);
    pthread_cond_broadcast(&condA);
    pthread_cond_broadcast(&condS);
    pthread_mutex_unlock(&mutex);

    pthread_exit(NULL);
}

int main(){
    createList(&p);
    printList(p);
    pthread_t threadB;
    pthread_create(&threadB, NULL, B, (void*)0);
    pthread_t threadZ;
    pthread_create(&threadZ, NULL, Z, (void*)1);
    pthread_t threadD;
    pthread_create(&threadD, NULL, D, (void*)2);
    pthread_t threadO;
    pthread_create(&threadO, NULL, O, (void*)3);
   
    pthread_exit(NULL);
}