#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>

#define BROJ_ZAPOSLENIH 10
#define BROJ_PROGRAMERA 10	

#define X 3

typedef struct programer
{
	int id;
	int brojProjekata;
	struct programer *sledeci;
}Programer;

pthread_mutex_t mutexPretraga;
pthread_mutex_t mutexDodavanje;
pthread_mutex_t mutex;
pthread_cond_t cond;
pthread_cond_t condRadiSeDodavanje;

int radiSeBrisanje = 0; //0-Ne, 1-DA
int radiSeDodavanje = 0; //0-Ne, 1-DA
int radiSeDodavanjeProjekata = 0; //0-Ne, 1-DA
int radiSePretraga = 0; //0-Ne, 1-DA

Programer *programerList;

void *funcBX(void *t) //odgovara na pitanje da li u kompaniji trenutno radi X ili vise programera
{
	long tId = (long)t;

	pthread_mutex_lock(&mutexPretraga);
	radiSePretraga = 1;
	while (radiSeBrisanje)
		pthread_cond_wait(&cond, &mutexPretraga);

	pthread_mutex_unlock(&mutexPretraga);

	int brojKojihNemajuProjekat = 0;

	Programer *temp = programerList;
	while(temp != NULL){
		if (temp->brojProjekata == 0)
			brojKojihNemajuProjekat++;
		temp = temp->sledeci;
	} 

	pthread_mutex_lock(&mutexPretraga);
	radiSePretraga = 0;
	pthread_mutex_unlock(&mutexPretraga);

	if(brojKojihNemajuProjekat > X)
		printf("FuncBX: Broj programera koji nemaju projekat je veci od X i on iznosi: %d\n", brojKojihNemajuProjekat);
	else
		printf("FuncBX: Broj programera koji nemaju projekat je manji od X i on iznosi: %d\n", brojKojihNemajuProjekat);

	pthread_exit(NULL);
}

void *funcZ(void *t) //zaposaljavanje novog programera
{
	long tId = (long)t;

	pthread_mutex_lock(&mutex);
	radiSeDodavanje = 1;
	while (radiSeBrisanje || radiSeDodavanje)
		pthread_cond_wait(&cond, &mutex);
	pthread_mutex_unlock(&mutex);
	Programer *novi = (Programer*)malloc(sizeof(Programer));
	novi->id = tId;
	novi->brojProjekata = 0;
	novi->sledeci = programerList;
	programerList = novi;

	pthread_mutex_lock(&mutexPretraga);
	radiSeDodavanje = 0;
	pthread_cond_signal();
	pthread_mutex_unlock(&mutexPretraga);

	pthread_exit(NULL);
}

void *funcDIDP(void *t) //postavljanjeProgrameranaProjekatP
{
	long tId = (long)t;

	pthread_mutex_lock(&mutexDodavanje);
	
	while (radiSeBrisanje)
		pthread_cond_wait(&cond, &mutexDodavanje);

	Programer *temp = programerList;
	while(temp != NULL)
	{
		if(temp->id == tId)
		{
			temp->brojProjekata++;
		}
		temp = temp->sledeci;
	}

	pthread_mutex_unlock(&mutexDodavanje);

	pthread_exit(NULL);
}

void *funcO(void *t) //otpustanje programera
{
	long tId = (long)t;

	pthread_mutex_lock(&mutex);
	radiSeBrisanje = 1;

	Programer *pret = programerList;

	if(pret != NULL)
	{
		Programer *temp = pret->sledeci;
		while(temp != NULL)
		{
			if(temp->id == tId)
			{
				pret->sledeci = temp->sledeci;
				free(temp);
			}
			temp = temp->sledeci;
		}
	}
	
	radiSeBrisanje=0;
	pthread_cond_broadcast(&cond);

	pthread_mutex_unlock(&mutex);

	pthread_exit(NULL);
}


int main()
{
	pthread_t threads[BROJ_ZAPOSLENIH];
	
	pthread_mutex_init(&mutexPretraga, NULL);
	pthread_mutex_init(&mutex, NULL);
	pthread_mutex_init(&mutexDodavanje, NULL);
	pthread_cond_init(&cond, NULL);
	long i,j;

	for (i = 0; i < 5; ++i)
	{
		pthread_create(&threads[i], NULL, funcZ, (void*)i);
	}

	pthread_create(&threads[i], NULL, funcBX, (void*)i);
	pthread_create(&threads[++i], NULL, funcBX, (void*)i);
	pthread_create(&threads[++i], NULL, funcBX, (void*)i);
	pthread_create(&threads[++i], NULL, funcBX, (void*)i);
	pthread_create(&threads[++i], NULL, funcBX, (void*)i);

	//pthread_create(&threads[++i], NULL, funcO, (void*)1); //povecavam programeru 1 projekte

	pthread_create(&threads[++i], NULL, funcDIDP, (void*)2); //povecavam programeru 1 projekte


	for (i = 0; i < 11; ++i)
	{
		pthread_join(threads[i], NULL);
	}

	while(programerList != NULL){
		printf("Programer: %d %d\n", programerList->id, programerList->brojProjekata);
		programerList = programerList->sledeci;
	}


	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&cond);

	pthread_exit(NULL);
}