#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <pthread.h>

#define N 10
#define X 3
#define ID 1
#define P 2

#define PROVERI_BEZ_ZADATKA 0
#define ZAPOSLI 1
#define POSTAVI_NA_PROJEKAT 0
#define OTPUSTI 0
#define NEMA_PROJEKAT 0

typedef struct programer{
    long id;
    int brProj;
    struct programer* sledeci;
} Programer;

Programer *programeri = NULL;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
pthread_cond_t condMenja = PTHREAD_COND_INITIALIZER;
pthread_cond_t condCita = PTHREAD_COND_INITIALIZER;
pthread_cond_t condBrise = PTHREAD_COND_INITIALIZER;
pthread_cond_t condBriseUpisuje = PTHREAD_COND_INITIALIZER;

int cita = 0;
int upisuje = 0;
int brise = 0;
int brProgramera = 0;

void dodaj()
{
    Programer *novi = (Programer*) malloc(sizeof(Programer));

    novi->id = brProgramera;
    novi->brProj = NEMA_PROJEKAT;
    novi->sledeci = programeri;
    programeri = novi;

    brProgramera++;
}

void ispisi()
{
    Programer *pom = programeri;

    while (pom != NULL)
    {
        printf ("Programer %ld: Projekat %d\n", pom->id, pom->brProj);

        pom = pom->sledeci;
    }
}

void* track()
{
    while (true)
    {
        printf("Brise %d, upisuje %d, cita %d.\n", brise, upisuje, cita);
        sleep(1);
    }
}

void* proveriBezProjekta()
{
    int brojProgBezProj = 0;

    printf("Cekam PROVERU programera bez projekta.\n");
    pthread_mutex_lock(&mutex);
    while (brise != 0)
        pthread_cond_wait(&condMenja, &mutex);
    cita++;
    pthread_mutex_unlock(&mutex);

    {  // cita
        printf("Radim PROVERU programera bez projekta.\n");
        Programer *pom = programeri;

        while (pom->sledeci != NULL)
        {
            if (pom->brProj == NEMA_PROJEKAT)
            {
                brojProgBezProj++;
            }

            pom = pom->sledeci;
        }

        if (brojProgBezProj >= X)
        {
            printf("U kompaniji ima vise od %d(%d) programera koji nemaju projekat.", X, brojProgBezProj);
        }
        else 
        {
            printf("U kompaniji nema vise od %d(%d) programera koji nemaju projekat.", X, brojProgBezProj);
        }
        sleep(5);
    }

    pthread_mutex_lock(&mutex);
    cita--;
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);

    printf("Zavrsio PROVERU programera bez projekta.\n");
    
    pthread_exit(NULL);
}

void* zaposli()
{
    printf("Cekam ZAPOSLJENJE programera.\n");
    pthread_mutex_lock(&mutex);
    while (brise != 0 || upisuje != 0)
        pthread_cond_wait(&cond, &mutex);
    upisuje++;
    pthread_mutex_unlock(&mutex);

    { // upisuje
        printf("Radim ZAPOSLJENJE programera.\n");
        Programer *novi = (Programer*) malloc(sizeof(Programer));

        novi->id = brProgramera;
        novi->brProj = NEMA_PROJEKAT;
        novi->sledeci = programeri;
        programeri = novi;

        brProgramera++;
        // ispisi();
        sleep(5);
    }

    pthread_mutex_lock(&mutex);
    upisuje--;
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);
    
    printf("Zavrsio ZAPOSLJENJE programera.\n");
    pthread_exit(NULL);
}

void* postaviNaProjekat()
{
    printf("Cekam POSTAVLJANJE programera na projekat.\n");
    pthread_mutex_lock(&mutex);
    while (brise != 0 || upisuje != 0 || cita != 0)
        pthread_cond_wait(&cond, &mutex);
    upisuje++;
    pthread_mutex_unlock(&mutex);

    { // upisuje
        printf("Radim POSTAVLJANJE programera na projekat.\n");
        Programer *pom = programeri;

        while (pom != NULL)
        {
            if (pom->id == ID)
            {
                pom->brProj = P;
                break;
            }

            pom = pom->sledeci;
        }
        // ispisi();
        sleep(5);
    }
    
    pthread_mutex_lock(&mutex);
    upisuje--;
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);

    printf("Zavrsio POSTAVLJANJE programera na projekat.\n");
    pthread_exit(NULL);
}

void* otpusti()
{
    printf("Cekam OTPUST programera.\n");
    pthread_mutex_lock(&mutex);
    while (brise != 0 || upisuje != 0 || cita != 0)
        pthread_cond_wait(&cond, &mutex);
    brise++;
    pthread_mutex_unlock(&mutex);

    { // brise
        printf("Radim OTPUST programera.\n");
        Programer *pom = programeri, *preth = NULL;

        if (pom->sledeci == NULL)
        {
            if (pom->id == ID)
            {
                programeri = NULL;
            }
        }
        else 
        {
            while (pom->sledeci != NULL)
            {
                if (pom->id == ID)
                {
                    preth->sledeci = pom->sledeci;
                    break;
                }

                preth = pom;
                pom = pom->sledeci;
            }
        } 
        // ispisi();
        sleep(5);
    }   

    pthread_mutex_lock(&mutex);
    brise--;
    pthread_cond_broadcast(&cond);
    pthread_mutex_unlock(&mutex);

    printf("Zavrsio OTPUST programera.\n");
    pthread_exit(NULL);
}

int main()
{
    programeri = (Programer*) malloc(N * sizeof(Programer));

    dodaj();
    dodaj();
    dodaj();

    // ispisi();

    pthread_t adminProveri, adminZaposli, adminPostavi, adminOtpusti, tracker;

    pthread_create(&tracker, NULL, track, NULL);
    pthread_create(&adminProveri, NULL, proveriBezProjekta, NULL);
    pthread_create(&adminZaposli, NULL, zaposli, NULL);
    pthread_create(&adminPostavi, NULL, postaviNaProjekat, NULL);
    pthread_create(&adminOtpusti, NULL, otpusti, NULL);

    pthread_join(tracker, NULL);
    pthread_join(adminProveri, NULL);
    pthread_join(adminZaposli, NULL);
    pthread_join(adminPostavi, NULL);
    pthread_join(adminOtpusti, NULL);

    pthread_exit(NULL);

    return 0;
}